//*****************************************************************************
//              \ /
// No Copyright (c) 2003
//              / \
// No rights reserved. This material contains published,
// non-copyrighted work, which includes non-confidential and
// non-proprietary information of Kurt Santel.
//
// Source: Robot.cpp
//
// Description: Server-side component that keeps track of the properties
//              of a Robot
//
//*****************************************************************************

#include "Robot.h"
#include "Square.h"

#include <math.h>               // Needed for ::floor()

/////////////////////////////////////////////////////////////////////////////////////
// Static variables used to check if the sequence of the function calls is correct
static bool s_checkPreConditionsCalled = false;
static bool s_doTurnCalled = false;

/////////////////////////////////////////////////////////////////////////////////////
// Robot

Robot::Robot(IDType id, AttackType a, DefenseType d)
: m_hasTurn( false )
, m_hadTurn( false )
, m_registered( false )
, m_ID( id )
, m_position( NULL )
, m_attack( a )
, m_defense( d )
, m_food( valMaxFood )
, m_life( valMaxLife )
, m_wallHit( false )
, m_robotInSight( false )
, m_dead( false )
, m_attacked( false )
{
    m_heading = North;
}

Robot::~Robot(void)
{
    if( m_position )
    {
        m_position->removeRobot( this );    // Let the Square know that you died
    }
}

bool Robot::checkPreConditions(void)
{
    // Reset static check variables
    s_checkPreConditionsCalled = false;
    s_doTurnCalled = false;

#ifdef _DEBUG
    m_food = 100;
#endif

    if( (false == m_hasTurn) ||   // Does the robot have a turn?
        (false == m_registered) )
    {
        return( false );
    }

    s_checkPreConditionsCalled = true;

    return( true );
}

bool Robot::doTurn(skipTurnCommand const & skip)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    s_doTurnCalled = true;

    return( true );
}

bool Robot::doTurn(restCommand const & rest)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    // Get your life back together
    m_life += valRest;

    if( m_life > valMaxLife )
    {
        m_life = valMaxLife;
    }

    // Doing nothing makes you hungry
    m_food += valHunger;

    s_doTurnCalled = true;

    return( true );
}

bool Robot::doTurn(moveForwardCommand const & mf)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    Square * neighbour = NULL;

    switch( m_heading )
    {
    case North: neighbour = m_position->getNorth(); break;
    case East:  neighbour = m_position->getEast();  break;
    case South: neighbour = m_position->getSouth(); break;
    case West:  neighbour = m_position->getWest();  break;
    }

    if( NULL != neighbour )
    {
        m_position->removeRobot( this );    // Let the Square know that you moved
        neighbour->addRobot( this );

        m_position = neighbour;
    }
    else
    {
        m_wallHit = true;                   // We are hitting an outside wall
    }

    // Get tired
    m_life += valWalk;

    // Doing nothing makes you hungry
    m_food += valHunger;

    s_doTurnCalled = true;

    return( true );
}

bool Robot::doTurn(moveBackwardCommand const & mb)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    Square * neighbour = NULL;

    switch( m_heading )
    {
    case North: neighbour = m_position->getSouth(); break;
    case East:  neighbour = m_position->getWest();  break;
    case South: neighbour = m_position->getNorth(); break;
    case West:  neighbour = m_position->getEast();  break;
    }

    if( NULL != neighbour )
    {
        m_position->removeRobot( this );    // Let the Square know that you moved
        neighbour->addRobot( this );

        m_position = neighbour;
    }
    else
    {
        m_wallHit = true;                  // We are hitting an outside wall
    }

    // Get tired
    m_life += valWalk;

    // Doing nothing makes you hungry
    m_food += valHunger;

    s_doTurnCalled = true;

    return( true );
}

bool Robot::doTurn(turnLeftCommand const & tl)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    switch( m_heading )
    {
    case North: m_heading = West;  break;
    case East:  m_heading = North; break;
    case South: m_heading = East;  break;
    case West:  m_heading = South; break;
    }

    // Doing nothing makes you hungry
    m_food += valHunger;

    s_doTurnCalled = true;

    return( true );
}

bool Robot::doTurn(turnRightCommand const & tr)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    switch( m_heading )
    {
    case North: m_heading = East;  break;
    case East:  m_heading = South; break;
    case South: m_heading = West;  break;
    case West:  m_heading = North; break;
    }

    // Doing nothing makes you hungry
    m_food += valHunger;

    s_doTurnCalled = true;

    return( true );
}

bool Robot::doTurn(eatCommand const & eat)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    // Njammie
    if( Square::Food == m_position->getType() )
    {
        m_food += valEat;
        m_life += valRest;
    }

    s_doTurnCalled = true;

    return( true );
}

bool Robot::doTurn(fightCommand const & fight)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || s_doTurnCalled )
    {
        return( false );
    }

    // Find the next Square...
    Square * neighbour = NULL;

    switch( m_heading )
    {
    case North: neighbour = m_position->getNorth(); break;
    case East:  neighbour = m_position->getEast();  break;
    case South: neighbour = m_position->getSouth(); break;
    case West:  neighbour = m_position->getWest();  break;
    }

    if( neighbour )
    {
        std::list<Robot *> robotsInSight;

        neighbour->getRobots( robotsInSight );

        // ... and attack all the robots that are in sight
        for( std::list<Robot *>::iterator i=robotsInSight.begin(); i != robotsInSight.end(); ++i )
        {
            Robot * victim = *i;
            victim->m_life -= (LifeType)::floor( (m_attack + victim->m_defense) / victim->m_defense );
            victim->m_attacked = true;
        }
    }

    m_life += valAttacking;
    m_food += valHunger;        // Fighting makes you hungry

    s_doTurnCalled = true;

    return( true );
}

bool Robot::checkPostConditions(void)
{
    // Check sequence of function calls
    if( !s_checkPreConditionsCalled || !s_doTurnCalled )
    {
        return( false );
    }

    m_hasTurn = false;
    m_hadTurn = true;

    //*********************************************
    //     check some things like:
    //     - is the robot under fire?
    //     - is the robot eating?
    //     - ...
    //*********************************************

    // Am I starving?
    if( 0 >= m_food )
    {
        m_life += valStarve;
    }

    // Am I dead?
    if( 0 >= m_life )
    {
        m_dead = true;
    }
    else
    {
        m_dead = false;
    }

    // Have I found food?
    if( Square::Food == m_position->getType() )
    {
        m_foodFound = true;
    }
    else
    {
        m_foodFound = false;
    }

    // Is there a Robot in sight? (one Square further)
    Square * neighbour;

    switch( m_heading )
    {
    case North: neighbour = m_position->getNorth(); break;
    case East:  neighbour = m_position->getEast();  break;
    case South: neighbour = m_position->getSouth(); break;
    case West:  neighbour = m_position->getWest();  break;
    }

    if( NULL != neighbour )
    {
        std::list<Robot *> robotsInSight;

        neighbour->getRobots( robotsInSight );

        if( !robotsInSight.empty() )
        {
            m_robotInSight = true;
        }
        else
        {
            m_robotInSight = false;
        }
    }

    return( true );
}
